// RefactoringRename.h

#pragma once

#ifndef _REFACTORING_RENAME_H_
#define _REFACTORING_RENAME_H_

#include <deque>
#include <map>
#include <set>

#include "ICsFile.h"
#include "FastDeque.h"
#include "CsToken.h"
#include "AstVisitor.h"
#include "CsParserTypes.h"


struct CsCompilationUnit;
class ICsFile;

//------------------------------------------------------------------------------

enum entity_lookup_flag
{
	elf_base_was_lookup = 0x01,
	elf_derived_was_lookup = 0x02,
	elf_was_notified_from_base_class = 0x04,
	elf_was_notified_from_base_interface = 0x08,
	elf_was_notified_from_derived = 0x10,
};

struct entity_lookup_struct
{
	entity_lookup_struct()
		: flags(0)
	{}

	inline void setFlag ( entity_lookup_flag elf, bool flag )
	{
		if ( flag )
			flags |= elf;
		else
			flags &= ~elf;
	}
	inline bool isFlagSet ( entity_lookup_flag elf ) const
	{
		return ( flags & elf ) != 0;
	}
	void setBaseWasLookup ( bool flag ) { setFlag ( elf_base_was_lookup, flag ); }
	bool isBaseWasLookup() const { return isFlagSet ( elf_base_was_lookup ); }
	void setDerivedWasLookup ( bool flag ) { setFlag ( elf_derived_was_lookup, flag ); }
	bool isDerivedWasLookup() const { return isFlagSet ( elf_derived_was_lookup ); }
	void setWasNotifiedFromBaseClass ( bool flag ) { setFlag ( elf_was_notified_from_base_class, flag ); }
	bool isWasNotifiedFromBaseClass() const { return isFlagSet ( elf_was_notified_from_base_class ); }
	void setWasNotifiedFromBaseInterface ( bool flag ) { setFlag ( elf_was_notified_from_base_interface, flag ); }
	bool isWasNotifiedFromBaseInterface() const { return isFlagSet ( elf_was_notified_from_base_interface ); }
	void setWasNotifiedFromDerived ( bool flag ) { setFlag ( elf_was_notified_from_derived, flag ); }
	bool isWasNotifiedFromDerived() const { return isFlagSet ( elf_was_notified_from_derived ); }

	unsigned flags;
};

//------------------------------------------------------------------------------
// RefactoringRename
//------------------------------------------------------------------------------

class RefactoringRename
{
public:
	void setProjectData (
		CsEntityNamespace * global_namespace,
		const std::deque<ICsFile*> & files
	);

	void rename ( const char * path, unsigned line, const char * old_identifier, unsigned identifier_number_in_line, const char * new_identifier );
	void rename ( const char * path, unsigned begin, unsigned end, const char * new_name );
	void rename ( ICsFile * file, CsTokenIndex token_index );
	void initBeforeRefactoring();
	ICsFile * findFile ( const char * path );
	MediumFastDeque<CsToken>::iterator getToken ( ICsFile * file, unsigned begin, unsigned end );
	void renameLocalEntity ( CsEntity * entity );
	void renameVariableOrConstant ( CsEntity * entity );
	void renameMethod ( CsEntityMethod * entity );
	void collectTypeMethods (
		CsEntityType * type,
		CsEntityType * from_type,
		CsEntityMethod * from_method,
		bool bFromDerived );
	void renameProperty ( CsEntityProperty * entity );
	void collectTypeProperties (
		CsEntityType * type,
		CsEntityType * from_type,
		CsEntityProperty * from_property,
		bool bFromDerived );
	void renameEvent ( CsEntity * entity );
	void renameType ( CsEntityType * entity );
	CsTokenIndex evalCsTokenIndex (
		CsCompoundIdentifier & compound_identifier,
		CsEntity * entity,
		CsEntity * find_entity
	);
	void renameUsingDirectiveAlias ( CsUsingAliasDirective * ud );
	void renameNamespace ( CsEntityNamespace * entity );
	void renameLabel ( CsLabel * label );

	static CsFile * getFile ( CsNode * node );
	static CsNode * getMethod ( CsNode * node );
	void addTokenToReplace ( CsFile * file, CsTokenIndex token_index )
	{
		tokensToReplace_.push_back
			( std::pair<ICsFile*,CsTokenIndex> ( ICsFile::getICsFile(file), token_index ) );
	}
	void replaceTokens ( bool bLocalRename );

	CsEntityNamespace * global_namespace_;
	std::map<std::string,ICsFile*> files_;
	PHashString new_name_;
	MediumFastDeque<std::pair<ICsFile*,CsTokenIndex> > tokensToReplace_;
	typedef MediumFastDeque<std::pair<ICsFile*,CsTokenIndex> >::iterator token_to_rename_iterator;
	CsNode * method_node_;
	bool bAssemblyMethod_;
	bool bAssemblyProperty_;
	std::set<CsEntity*> method_set_;
	std::set<CsEntity*> property_set_;
	std::map<CsEntityType*,entity_lookup_struct> type_lookup_map_;
// reparsing
	bool bNeedGlobalEntityReparsing_;
};

//------------------------------------------------------------------------------
// CsAstVisitor1 
//------------------------------------------------------------------------------

class CsAstVisitor1 : public CsAstVisitor
{
public:
	virtual void visit ( CsUsingAliasDirective * node ) { visitNode ( node ); }
	virtual void visit ( CsUsingNamespaceDirective * node ) { visitNode ( node ); }
	virtual void visit ( CsTypeRef * node ) { visitNode ( node ); }
	virtual void visit ( CsSimpleName * node ) { visitNode ( node ); }
	virtual void visit ( CsLiteral * node ) { visitNode ( node ); }
	virtual void visit ( CsEmbeddedTypeMemberAccess * node ) { visitNode ( node ); }
	virtual void visit ( CsThisAccess * node ) { visitNode ( node ); }
	virtual void visit ( CsBaseMemberAccess * node ) { visitNode ( node ); }
	virtual void visit ( CsBaseIndexerAccess * node ) { visitNode ( node ); }
	virtual void visit ( CsTypeofExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsCheckedExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsUncheckedExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsLockStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsSizeofExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsParenthesizedExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsNewObjectExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsNewArrayExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsInvocationExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsElementAccess * node ) { visitNode ( node ); }
	virtual void visit ( CsMemberAccess * node ) { visitNode ( node ); }
	virtual void visit ( CsPointerMemberAccess * node ) { visitNode ( node ); }
	virtual void visit ( CsPostIncrementDecrementExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsBinaryExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsAsIsExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsConditionalExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsOperatorUnaryExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsPreIncrementDecrementExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsCastUnaryExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsAssignmentExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsExpressionList * node ) { visitNode ( node ); }
	virtual void visit ( CsNamedArgument * node ) { visitNode ( node ); }
	virtual void visit ( CsAttribute * node ) { visitNode ( node ); }
	virtual void visit ( CsAttributeSection * node ) { visitNode ( node ); }
	virtual void visit ( CsAttributes * node ) { visitNode ( node ); }
	virtual void visit ( CsTypeBase * node ) { visitNode ( node ); }
	virtual void visit ( CsClass * node ) { visitNode ( node ); }
	virtual void visit ( CsStruct * node ) { visitNode ( node ); }
	virtual void visit ( CsInterface * node ) { visitNode ( node ); }
	virtual void visit ( CsEnumMember * node ) { visitNode ( node ); }
	virtual void visit ( CsEnum * node ) { visitNode ( node ); }
	virtual void visit ( CsFormalParameter * node ) { visitNode ( node ); }
	virtual void visit ( CsFormalParameterList * node ) { visitNode ( node ); }
	virtual void visit ( CsDelegate * node ) { visitNode ( node ); }
	virtual void visit ( CsNamespace * node ) { visitNode ( node ); }
	virtual void visit ( CsArgument * node ) { visitNode ( node ); }
	virtual void visit ( CsArgumentList * node ) { visitNode ( node ); }
	virtual void visit ( CsBlock * node ) { visitNode ( node ); }
	virtual void visit ( CsLabel * node ) { visitNode ( node ); }
	virtual void visit ( CsLabeledStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsEmptyStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsExpressionStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsDeclarationStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsLocalConstantDeclarator * node ) { visitNode ( node ); }
	virtual void visit ( CsLocalConstantDeclaration * node ) { visitNode ( node ); }
	virtual void visit ( CsIfStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsSwitchLabel * node ) { visitNode ( node ); }
	virtual void visit ( CsSwitchSection * node ) { visitNode ( node ); }
	virtual void visit ( CsSwitchStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsWhileStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsDoWhileStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsLocalVariableDeclarator * node ) { visitNode ( node ); }
	virtual void visit ( CsLocalVariableDeclaration * node ) { visitNode ( node ); }
	virtual void visit ( CsStatementExpressionList * node ) { visitNode ( node ); }
	virtual void visit ( CsForStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsForeachStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsBreakStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsContinueStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsGotoStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsReturnStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsThrowStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsCatchClause * node ) { visitNode ( node ); }
	virtual void visit ( CsTryStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsCheckedStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsUncheckedStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsUsingStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsUnsafeStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsFixedPointDeclarator * node ) { visitNode ( node ); }
	virtual void visit ( CsFixedStatement * node ) { visitNode ( node ); }
	virtual void visit ( CsStackAllocExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsConstructorDeclaration * node ) { visitNode ( node ); }
	virtual void visit ( CsConstantDeclarator * node ) { visitNode ( node ); }
	virtual void visit ( CsConstantDeclaration * node ) { visitNode ( node ); }
	virtual void visit ( CsVariableDeclarator * node ) { visitNode ( node ); }
	virtual void visit ( CsVariableDeclaration * node ) { visitNode ( node ); }
	virtual void visit ( CsArrayInitializer * node ) { visitNode ( node ); }
	virtual void visit ( CsMethod * node ) { visitNode ( node ); }
	virtual void visit ( CsPropertyAccessor * node ) { visitNode ( node ); }
	virtual void visit ( CsProperty * node ) { visitNode ( node ); }
	virtual void visit ( CsEventAccessor * node ) { visitNode ( node ); }
	virtual void visit ( CsEventDeclarator * node ) { visitNode ( node ); }
	virtual void visit ( CsEvent * node ) { visitNode ( node ); }
	virtual void visit ( CsIndexer * node ) { visitNode ( node ); }
	virtual void visit ( CsOperator * node ) { visitNode ( node ); }
	virtual void visit ( CsConversionOperatorDeclaration * node ) { visitNode ( node ); }
	virtual void visit ( CsDestructorDeclaration * node ) { visitNode ( node ); }
	virtual void visit ( CsCompilationUnit * node ) { visitNode ( node ); }
	virtual void visit ( CsMakeRefExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsRefTypeExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsRefValueExpression * node ) { visitNode ( node ); }
	virtual void visit ( CsArgListExpression * node ) { visitNode ( node ); }

	virtual void visitNode ( CsNode * node ) =0;
};

//------------------------------------------------------------------------------
// CsEntityVisitor1
//------------------------------------------------------------------------------

class CsEntityVisitor1 : public CsEntityVisitor
{
public:
	//virtual void visit ( CsEntityGlobalNamespace * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityNamespace * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityBlock * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityFormalParameter * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityFormalParameterList * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityMethod * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityClass * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityStruct * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityInterface * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityEnum * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityConstant * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityVariable * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityBlockVariable * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityDelegate * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityProperty * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityEvent * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityLocalConstant * entity ) { visitEntity ( entity ); }
	virtual void visit ( CsEntityLocalVariable * entity ) { visitEntity ( entity ); }

	virtual void visitEntity ( CsEntity * entity ) =0;
};

//------------------------------------------------------------------------------
// CsFindTokenVisitor
//------------------------------------------------------------------------------

class CsFindTokenVisitor : public CsAstVisitor1
{
public:
	virtual void visitNode ( CsNode * node );

	CsNode * findNodeByTokenIndex ( CsNode * startNode, CsTokenIndex index );

	CsTokenIndex tmp_token_index;
	CsTokenCollector tmp_tokens;
	CsNode * tmp_found_node;
};

//------------------------------------------------------------------------------
// CsNodeCollectorVisitor
//------------------------------------------------------------------------------

class CsNodeCollectorVisitor : public CsAstVisitor1
{
public:
	CsNodeCollectorVisitor();
	void init();
	
	virtual void visitNode ( CsNode * node );
	void collectNodes ( CsNode * startNode );
	bool isEntityExist (
		CsCompoundIdentifier & compound_identifier,
		CsEntity * entity,
		bool bSkipFirstEntity = false
	);

	MediumFastDeque<CsNode*> nodes_;
	bool collect_node_flags_[n_end];
	bool bCheckEntities_;
	std::set<CsEntity*> entities_;
	bool bCheckCompoundIdentifier_;
	bool bCheckUsingAliasDirective_;
	CsUsingAliasDirective * using_alias_directive_;
	bool bProjectEntities_;
};

//------------------------------------------------------------------------------
// CsEntityCollectorVisitor
//------------------------------------------------------------------------------

class CsEntityCollectorVisitor : public CsEntityVisitor1
{
public:
	CsEntityCollectorVisitor();
	void init();

	virtual void visitEntity ( CsEntity * entity );
	void collectEntities ( CsEntity * startEntity );

	MediumFastDeque<CsEntity*> entities;
	bool collect_entities_flags[ent_end];
	bool bCollectDerivedTypes;
	CsEntityType * base_type;
};

//------------------------------------------------------------------------------


#endif // _REFACTORING_RENAME_H_